import inspect
import json
import logging
import os
from pathlib import Path
from pydantic import BaseModel, Field
from typing import Any, Callable, Dict, List, Literal, Optional, cast

logger = logging.getLogger(__name__)

DEBUG_DIR = "debug"
TESTS_DIR = "tests"
RESULTS_FILENAME = "result.json"
API_PARAM = "api"
ARGS_PARAM = "args"

class ToolInfo(BaseModel):
	name: str
	description: str
	parameters: Any
	signature: str
	full_description:str

	@classmethod
	def from_function(cls, fn: Callable) -> "ToolInfo":
		# Assumes @tool decorator from langchain_core https://python.langchain.com/docs/how_to/custom_tools/
		# or a plain function with doc string
		def doc_summary(doc:str): 
			paragraphs = [p.strip() for p in doc.split("\n\n") if p.strip()]
			return paragraphs[0] if paragraphs else ""
		
		fn_name = fn.name if hasattr(fn, 'name') else fn.__name__ # type: ignore
		sig =fn_name + str(inspect.signature(fn))
		full_desc = fn.description if hasattr(fn,'description') else fn.__doc__.strip() if fn.__doc__ else (inspect.getdoc(fn) or "") # type: ignore
		return cls(
            name=fn_name,
			description=doc_summary(full_desc),
			full_description = full_desc,
            parameters=fn.args_schema.model_json_schema() if hasattr(fn, 'args_schema') else inspect.getdoc(fn), # type: ignore
			signature=sig,
        )
    

class FileTwin(BaseModel):
    file_name: str
    content: str

    def save(self, folder: str) -> "FileTwin":
        full_path = os.path.join(folder, self.file_name)
        parent = Path(full_path).parent
        os.makedirs(parent, exist_ok=True)
        with open(full_path, "w") as file:
            file.write(self.content)
        return self

    def save_as(self, folder: str, file_name: str) -> "FileTwin":
        file_path = os.path.join(folder, file_name)
        with open(file_path, "w") as file:
            file.write(self.content)
        return FileTwin(file_name=file_name, content=self.content)

    @staticmethod
    def load_from(folder: str, file_path: str) -> "FileTwin":
        with open(os.path.join(folder, file_path), "r") as file:
            data = file.read()
            return FileTwin(file_name=file_path, content=data)
        
        
class ToolGuardSpecItem(BaseModel):
    name: str = Field(..., description="Policy item name")
    description: str = Field(..., description="Policy item description")
    references: List[str] = Field(..., description="original texts")
    compliance_examples: Optional[List[str]] = Field(
        ..., description="Example of cases that comply with the policy"
    )
    violation_examples: Optional[List[str]] = Field(
        ..., description="Example of cases that violate the policy"
    )
    skip: bool = False

    def to_md_bulltets(self, items: List[str]) -> str:
        s = ""
        for item in items:
            s += f"* {item}\n"
        return s

    def __str__(self) -> str:
        s = "#### Policy item " + self.name + "\n"
        s += f"{self.description}\n"
        if self.compliance_examples:
            s += f"##### Positive examples\n{self.to_md_bulltets(self.compliance_examples)}"
        if self.violation_examples:
            s += f"##### Negative examples\n{self.to_md_bulltets(self.violation_examples)}"
        return s


class ToolGuardSpec(BaseModel):
    tool_name: str = Field(..., description="Name of the tool")
    policy_items: List[ToolGuardSpecItem] = Field(
        ...,
        description="Policy items. All (And logic) policy items must hold whehn invoking the tool.",
    )


def load_tool_policy(file_path: str, tool_name: str) -> ToolGuardSpec:
    with open(file_path, "r") as file:
        d = json.load(file)
    d["tool_name"] = tool_name
    spec = ToolGuardSpec.model_validate(d) #load deep
    return spec

class Domain(BaseModel):
    app_name: str = Field(..., description="Application name")
    toolguard_common: FileTwin = Field(
        ..., description="Pydantic data types used by toolguard framework."
    )
    app_types: FileTwin = Field(
        ..., description="Data types defined used in the application API as payloads."
    )
    app_api_class_name: str = Field(..., description="Name of the API class name.")
    app_api: FileTwin = Field(
        ..., description="Python class (abstract) containing all the API signatures."
    )
    app_api_size: int = Field(..., description="Number of functions in the API")


class RuntimeDomain(Domain):
    app_api_impl_class_name: str = Field(
        ..., description="Python class (implementaton) class name."
    )
    app_api_impl: FileTwin = Field(
        ..., description="Python class containing all the API method implementations."
    )

    def get_definitions_only(self):
        return Domain.model_validate(self.model_dump())


class PolicyViolationException(Exception):
    _msg: str

    def __init__(self, message: str):
        super().__init__(message)
        self._msg = message

    @property
    def message(self):
        return self._msg




import logging


def add_log_file_handler(log_file:str):
	formatter = logging.Formatter(
		fmt="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
		datefmt="%Y-%m-%d %H:%M:%S"
	)
	file_handler = logging.FileHandler(log_file, encoding="utf-8")
	file_handler.setFormatter(formatter)
	logging.getLogger().addHandler(file_handler)

def init_logging():
	logging.getLogger().setLevel(logging.INFO) #Default for other libraries
	logging.getLogger("toolguard").setLevel(logging.DEBUG) # debug for our library
	logging.getLogger("mellea").setLevel(logging.DEBUG)
	init_log_console_handler()

def init_log_console_handler():
	formatter = logging.Formatter(
		fmt="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
		datefmt="%Y-%m-%d %H:%M:%S"
	)

	# Set up console handler
	console_handler = logging.StreamHandler()
	console_handler.setFormatter(formatter)

	# Set up the root logger
	logging.basicConfig(
		level=logging.INFO,
		handlers=[console_handler]
	)




import os
import inspect
from typing import Callable
import sys
from pathlib import Path
from contextlib import contextmanager

from .str import to_snake_case

def py_extension(filename: str) -> str:
    return filename if filename.endswith(".py") else filename + ".py"


def un_py_extension(filename: str) -> str:
    return filename.removesuffix(".py") if filename.endswith(".py") else filename


def path_to_module(file_path: str) -> str:
    assert file_path
    parts = file_path.split("/")
    if parts[-1].endswith(".py"):
        parts[-1] = un_py_extension(parts[-1])
    return ".".join([to_snake_case(part) for part in parts])


def module_to_path(module: str) -> str:
    parts = module.split(".")
    return os.path.join(*parts[:-1], py_extension(parts[-1]))

@contextmanager
def temp_python_path(path: str):
    path = str(Path(path).resolve())
    if path not in sys.path:
        sys.path.insert(0, path)
        try:
            yield
        finally:
            sys.path.remove(path)
    else:
        # Already in sys.path, no need to remove
        yield

def extract_docstr_args(func:Callable) -> str:
    doc = inspect.getdoc(func)
    if not doc:
        return ""

    lines = doc.splitlines()
    args_start = None
    for i, line in enumerate(lines):
        if line.strip().lower() == "args:":
            args_start = i
            break

    if args_start is None:
        return ""

    # List of known docstring section headers
    next_sections = {
        "returns:",
        "raises:",
        "examples:",
        "notes:",
        "attributes:",
        "yields:",
    }

    # Capture lines after "Args:" that are indented
    args_lines = []
    for line in lines[args_start + 1 :]:
        # Stop if we hit a new section (like "Returns:", "Raises:", etc.)
        stripped = line.strip().lower()
        if stripped in next_sections:
            break
        args_lines.append(" " * 8 + line.strip())

    # Join all lines into a single string
    if not args_lines:
        return ""

    return "\n".join(args_lines)

def get_func_signature(obj)->str:
	if inspect.isfunction(obj):
		return inspect.signature(obj)
	if hasattr(obj, "func") and inspect.isfunction(obj.func): # a @tool
		return inspect.signature(obj.func)
	if hasattr(obj, "args_schema"):
		schema = obj.args_schema
		fields = schema.model_fields
		params = ", ".join(
			f"{name}: {field.annotation.__name__ if hasattr(field.annotation, '__name__') else field.annotation}"
			for name, field in fields.items()
		)
		return f"({params})"
	return None


from enum import Enum
from pydantic import BaseModel, Field, HttpUrl
from typing import List, Dict, Optional, Any, TypeVar, Union
import json
import yaml

from .array import not_none
from .dict import find_ref
from .http import MEDIA_TYPE_APP_JSON
from .jschema import JSchema
from .ref import Reference


class Contact(BaseModel):
    name: Optional[str] = None
    url: Optional[HttpUrl] = None
    email: Optional[str] = None


class License(BaseModel):
    name: str
    identifier: Optional[str] = None
    url: Optional[HttpUrl] = None


class Info(BaseModel):
    title: str
    summary: Optional[str] = None
    description: Optional[str] = None
    termsOfService: Optional[HttpUrl] = None
    contact: Optional[Contact] = None
    license: Optional[License] = None
    version: str


class ServerVariable(BaseModel):
    enum: Optional[List[str]] = None
    default: str
    description: Optional[str] = None


class Server(BaseModel):
    url: str
    description: Optional[str] = None
    variables: Optional[Dict[str, ServerVariable]] = None


class ExternalDocumentation(BaseModel):
    description: Optional[str] = None
    url: HttpUrl


class Tag(BaseModel):
    name: str
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentation] = None


class MediaType(BaseModel):
    schema_: Optional[Union[Reference, JSchema]] = Field(None, alias="schema")
    example: Optional[Any] = None
    examples: Optional[Dict[str, Any]] = None


class RequestBody(BaseModel):
    description: Optional[str] = None
    required: Optional[bool] = None
    content: Optional[Dict[str, MediaType]] = None

    @property
    def content_json(self):
        if self.content:
            return self.content.get(MEDIA_TYPE_APP_JSON)


class Response(BaseModel):
    description: Optional[str] = None
    content: Optional[Dict[str, MediaType]] = None

    @property
    def content_json(self):
        if self.content:
            return self.content.get(MEDIA_TYPE_APP_JSON)


class StrEnum(str, Enum):
    """An abstract base class for string-based enums."""

    pass


class ParameterIn(StrEnum):
    query = "query"
    header = "header"
    cookie = "cookie"
    path = "path"


class Parameter(BaseModel):
    name: str
    description: Optional[str] = None
    in_: ParameterIn = Field(ParameterIn.query, alias="in")
    required: Optional[bool] = None
    schema_: Optional[Union[Reference, JSchema]] = Field(None, alias="schema")


class Operation(BaseModel):
    summary: Optional[str] = None
    description: Optional[str] = None
    operationId: Optional[str] = None
    tags: Optional[List[str]] = None
    parameters: Optional[List[Union[Reference, Parameter]]] = None
    requestBody: Optional[Union[Reference, RequestBody]] = None
    responses: Optional[Dict[str, Union[Reference, Response]]] = None
    security: Optional[Dict[str, List[str]]] = None


class PathItem(BaseModel):
    summary: Optional[str] = None
    description: Optional[str] = None
    servers: Optional[List[Server]] = None
    parameters: Optional[List[Union[Reference, Parameter]]] = None
    get: Optional[Operation] = None
    put: Optional[Operation] = None
    post: Optional[Operation] = None
    delete: Optional[Operation] = None
    options: Optional[Operation] = None
    head: Optional[Operation] = None
    patch: Optional[Operation] = None
    trace: Optional[Operation] = None

    @property
    def operations(self):
        d = {
            "get": self.get,
            "put": self.put,
            "post": self.post,
            "delete": self.delete,
            "options": self.options,
            "head": self.head,
            "patch": self.patch,
            "trace": self.trace,
        }
        return {k: v for k, v in d.items() if v is not None}


class Components(BaseModel):
    schemas: Optional[Dict[str, JSchema]] = None
    responses: Optional[Dict[str, Response]] = None
    parameters: Optional[Dict[str, Parameter]] = None
    examples: Optional[Dict[str, Any]] = None
    requestBodies: Optional[Dict[str, RequestBody]] = None
    headers: Optional[Dict[str, Any]] = None
    securitySchemes: Optional[Dict[str, Any]] = None
    links: Optional[Dict[str, Any]] = None
    callbacks: Optional[Dict[str, Any]] = None
    pathItems: Optional[Dict[str, PathItem]] = None


BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


class OpenAPI(BaseModel):
    openapi: str = Field(..., pattern=r"^3\.\d\.\d+(-.+)?$")
    info: Info
    jsonSchemaDialect: Optional[HttpUrl] = (
        "https://spec.openapis.org/oas/3.1/dialect/WORK-IN-PROGRESS"
    )
    servers: Optional[List[Server]] = [Server(url="/")]
    paths: Dict[str, Union[Reference, PathItem]] = {}
    webhooks: Optional[Dict[str, PathItem]] = None
    components: Optional[Components] = None
    security: Optional[List[Dict[str, List[str]]]] = None  # Refined to List of Dicts
    tags: Optional[List[Tag]] = None
    externalDocs: Optional[ExternalDocumentation] = None

    def get_operation_by_operationId(self, operationId: str) -> Operation | None:
        for path_item in self.paths.values():
            for op in path_item.operations.values():
                if op.operationId == operationId:
                    return op

    def resolve_ref(
        self, obj: Reference | BaseModelT | None, object_type: type[BaseModelT]
    ) -> BaseModelT | None:
        if isinstance(obj, Reference):
            tmp = find_ref(self.model_dump(), obj.ref)
            return object_type.model_validate(tmp)
        return obj

    def save(self, file_name: str):
        if file_name.endswith(".json"):
            with open(file_name, "w", encoding="utf-8") as f:
                f.write(
                    self.model_dump_json(indent=2, by_alias=True, exclude_none=True)
                )
            return
        # TODO yaml
        raise NotImplementedError()


def read_openapi(file_path: str) -> OpenAPI:
    with open(file_path, "r") as file:
        if file_path.endswith("json"):
            d = json.load(file)
        else:
            d = yaml.safe_load(file)
    return OpenAPI.model_validate(d, strict=False)
